/*
 * PS02_to_PS04
 *
 * Copyright (C) 2003-2004 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <libzoom.h>

int verbose = 0;
int ptns = 0;

static void usage(char *prgname)
{
    printf("ps02_to_ps04 v1.00 Copyright (C) 2003-2004 Dwight Engen\n\n");
    printf("This program is free software; you can redistribute it and/or modify\n"
	   "it under the terms of the GNU General Public License as published by\n"
	   "the Free Software Foundation; either version 2 of the License, or\n"
	   "(at your option) any later version.\n\n");

    printf("Convert a Zoom PS02 Songs.pts to a PS04 RhythmXX.seq\n");
    printf("Usage: %s [OPTION] <song number> <file.pts> <output.seq>\n\n", prgname);
    printf(" -h  --help\n");
    printf(" -v  --verbose\n\n");

    exit(EXIT_FAILURE);
}



#if 0
static const char *PS04_event_name(song_event_type_t event)
{
    static const char event_name[][9] =
    {
	"Invalid",
	"Pattern",
	"Note",
	"Tempo",
	"Timesig",
	"Drum Vol",
	"Bass Vol",
	"Drum Kit",
	"Bass Kit",
    };
    if (event < 9)
	return event_name[event];
    else
	return "Unknown";
}



static void PS04_song_dump(ps04_song_t *song, ps04_seq_t  *seq)
{
    int i;
    int ppq;

    printf("Song    : %-8.8s\n", song->name);
    printf("Events  : %d\n", song->events);

    ppq = 0;
    for (i = 0; i < song->events; i++)
    {
	unsigned short data = song->event[i].data;

	printf("Event:%4.4d PPQ:%4.4d (%5.5d) Meas:%3.3d Type:%-10s ",
		i+1, song->event[i].next_ppq, ppq, ppq/(PS04_PPQ*4)+1,
		PS04_event_name(song->event[i].type));

	switch(song->event[i].type)
	{
	case SONG_EVENT_PATTERN:
	    printf("%8.8s (0x%3.3x)\n", seq->pattern[data]->name, data);
	    break;
	case SONG_EVENT_NOTE:
	    printf("%-2s%3s\n", note_root_name[data & 0x0f],
		   note_tone_name[(data & 0xf0) >> 4]);
	    break;

	case SONG_EVENT_TEMPO:
	    printf("%d.%d\n", data/10, data - ((data/10)*10));
	    break;

	case SONG_EVENT_TIMESIG:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_DRUM_VOL:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_BASS_VOL:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_DRUM_KIT:
	    break;
	case SONG_EVENT_BASS_KIT:
	    break;
	}

	/* Handle large/small PPQ's */
	if (song->event[i].next_ppq == 0xff)
	    ppq += *(uint32_t *)&song->event[++i];
	else
	    ppq += song->event[i].next_ppq;
    }
    printf("\n");
}
#endif


int main(int argc, char *argv[])
{
    ps02_pts_t	*pts;
    ps04_seq_t	*seq;
    int		 songnr;
    int		 opt;
    int		 i;

    static struct option long_options[] =
    {
	{"help",		0, 0, '?'},
	{"verbose",		0, 0, 'v'},

	{0,0,0,0},
    };

    if (argc < 4)
	usage(argv[0]);

    while ((opt = getopt_long(argc, argv, "v", long_options, NULL)) != -1)
    {
	switch(opt)
	{
	default:
	case '?':
	    usage(argv[0]);
	    break;
	case 'v':
	    verbose = 1;
	    break;
	}
    }

    if (argc - optind != 3)
	usage(argv[0]);

    songnr = strtoul(argv[optind++], NULL, 0);
    if (songnr > 100 || songnr == 0)
    {
	printf("Invalid song %d (1 - 100 are valid)\n", songnr);
	exit(EXIT_FAILURE);
    }

    pts = PS02_pts_open(argv[optind]);
    if (pts == NULL)
    {
	fprintf(stderr, "Unable to open %s\n", argv[optind]);
	exit(EXIT_FAILURE);
    }
    optind++;

    seq = PS04_seq_new();
    seq->F = fopen(argv[optind], "w+b");
    if (seq->F == NULL)
    {
	fprintf(stderr, "Unable to open %s\n", argv[optind]);
	exit(EXIT_FAILURE);
    }

    if (verbose)
    {
	ps02_song_t *song02;

	song02 = pts->song[songnr-1];
	printf("Song    : %-6.6s\n", song02->name);
	printf("Tempo   : %d\n", song02->tempo);

	for(i =  0; i < pts->rhy_pattern_cnt; i++)
	    printf("Pattern : %6.6s\n", pts->rhy_pattern[i]->name);

	for(i = 0; i < song02->measures; i++)
	    printf("Measure : %3.3d pattern:%6.6s\n", i+1,
		   pts->rhy_pattern[song02->measure[i].pattern]->name);
    }

    PS02_to_PS04(pts, seq, songnr-1);
    //PS04_song_dump(seq->song[0], seq);

    PS04_seq_write(seq);
    PS04_seq_close(seq);
    PS02_pts_close(pts);
    return EXIT_SUCCESS;
}
