/*
 * PS02 - Manage the Songs.pts file from a Zoom PS-02.
 *
 * Copyright (C) 2003-2004 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include "ps02.h"
#include <libzoom.h>

enum 
{
    SONG_COL_NUMBER,
    SONG_COL_NUMBER_TEXT,
    SONG_COL_NAME,
    SONG_COLS
};

enum 
{
    PATTERN_COL_NUMBER,
    PATTERN_COL_NUMBER_TEXT,
    PATTERN_COL_NAME,
    PATTERN_COLS
};

enum 
{
    PATCH_COL_NUMBER,
    PATCH_COL_NUMBER_TEXT,
    PATCH_COL_NAME,
    PATCH_COLS
};

enum
{
    RHY_COL_MEASURE_NUMBER,
    RHY_COL_MEASURE,
    RHY_COL_PATTERN,
    RHY_COL_BEATS
};

ps02_pts_t	*pts = NULL;
ps02_pch_t	*pch_file[2] = {NULL, NULL};		/* Preset, User */
ps02_song_t		*clipsong = NULL;
patch_t		 clippch;
GtkWidget	*MainWindow = NULL;
gint		 mcpy_beg;
gint		 mcpy_end;
gint		 mcpy_dst;
gboolean	 ignore_changed = FALSE;

static void	 modal_filesel(const char *title, char *file);
static gboolean	 modal_ask(gboolean can_cancel, const char *title,
			   const char *ok, const char *cancel,
			   const char *fmt,...);
static void	*gui(const gchar * widget_name);
static void	 open_files(const gchar *file);

void		 measures_populate(void);

void gui_sync(void)
{
    while(g_main_iteration(FALSE))
	;
}

void
gui_window_title (GtkWidget *win, const char *fmt,...)
{
    gchar title[256];
    va_list v_lst;

    va_start(v_lst, fmt);
    vsnprintf(title, sizeof(title)-1, fmt, v_lst);
    va_end(v_lst);
    gtk_window_set_title(GTK_WINDOW(win), title);
}



/* Patch handling */

void patch_select(gint pch_file_indx, gint pch)
{
    GtkTreeView		*tv;
    GtkTreePath		*path;
    GtkTreeModel	*model;
    GtkTreeIter		 row;

    if (pch_file_indx == 0)
	tv = gui("PresetPatchTreeView");
    else
	tv = gui("UserPatchTreeView");
    model = gtk_tree_view_get_model(tv);
    path = gtk_tree_path_new_from_indices(pch, -1);
    gtk_tree_model_get_iter(model, &row, path);
    gtk_tree_view_set_cursor(tv, path, NULL, FALSE);
    gtk_tree_path_free(path);
}

gboolean patch_selected_get(gint *pch_file_indx, gint *pch)
{
    GtkTreeModel	*model;
    GtkTreeIter		 row1;

    if (gtk_tree_selection_get_selected(
	   gtk_tree_view_get_selection(gui("PresetPatchTreeView")),
				       &model, &row1))
    {
	*pch_file_indx = 0;
    } else

    if (gtk_tree_selection_get_selected(
	   gtk_tree_view_get_selection(gui("UserPatchTreeView")),
				       &model, &row1))
    {
	*pch_file_indx = 1;
    } else
	return FALSE;


    gtk_tree_model_get(model, &row1, PATCH_COL_NUMBER, pch, -1);
    return TRUE;
}

void patch_update(gint pch_file_indx, gint pch)
{
    GtkTreeView		*tv;
    GtkListStore	*ls;
    GtkTreePath		*path;
    GtkTreeIter		 row;
    char		 buf[10];

    if (pch_file_indx == 0)
	tv = gui("PresetPatchTreeView");
    else
	tv = gui("UserPatchTreeView");

    ls = GTK_LIST_STORE(gtk_tree_view_get_model(tv));
    path = gtk_tree_path_new_from_indices(pch, -1);
    gtk_tree_model_get_iter(GTK_TREE_MODEL(ls), &row, path);

    gtk_list_store_set(ls, &row, PATCH_COL_NUMBER, pch, -1);
    sprintf(buf, "%2.2d", pch+1);
    gtk_list_store_set(ls, &row, SONG_COL_NUMBER_TEXT, buf, -1);
    sprintf(buf, "%-6.6s", pch_file[pch_file_indx]->pch[pch].name);
    gtk_list_store_set(ls, &row, SONG_COL_NAME, buf, -1);
    gtk_tree_path_free(path);
}

void patch_move(gint direction)
{
    gint		 pch_file_indx;
    gint		 pch1;
    gint		 pch2;
    patch_t		 temp;

    if (!patch_selected_get(&pch_file_indx, &pch1))
	return;

    pch2 = pch1 + direction;
    if (pch2 < 0 || pch2 >= NUM_PATCHES)
	return;

    /* Swap patches */
    temp = pch_file[pch_file_indx]->pch[pch1];
    pch_file[pch_file_indx]->pch[pch1] = pch_file[pch_file_indx]->pch[pch2];
    pch_file[pch_file_indx]->pch[pch2] = temp;
    pch_file[pch_file_indx]->modified = TRUE;

    /* Update GUI */
    patch_update(pch_file_indx, pch1);
    patch_update(pch_file_indx, pch2);
    patch_select(pch_file_indx, pch2);
    return;
}

void patch_selected(GtkTreeSelection *sel, gpointer user_data)
{
    GtkTreeSelection	*unsel;
    GtkTreeModel	*model;
    GtkTreeIter		 row1;
    int			 pch_file_indx = (int)user_data;

    /* Ignore the deselection changed signal */
    if (!gtk_tree_selection_get_selected(sel, &model, &row1))
	return;

    /* Deselect any selected patches in "other" patch list */
    if (pch_file_indx != 0)
	unsel = gtk_tree_view_get_selection(gui("PresetPatchTreeView"));
    else
	unsel = gtk_tree_view_get_selection(gui("UserPatchTreeView"));
    gtk_tree_selection_unselect_all(unsel);
}

void patch_list_populate(int pch_file_indx)
{
    int			 i;
    GtkListStore	*ls;
    GtkTreeView		*tv;
    GtkTreeIter		 row;
    GtkTreeViewColumn	*col;
    GtkTreeSelection	*sel;
    GtkCellRenderer	*rend;

    if (pch_file_indx == 0)
	tv = gui("PresetPatchTreeView");
    else
	tv = gui("UserPatchTreeView");

    ls = GTK_LIST_STORE(gtk_tree_view_get_model(tv));
    if (ls == NULL)
    {
	ls = gtk_list_store_new(PATCH_COLS, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING);
	gtk_tree_view_set_model(tv, GTK_TREE_MODEL(ls));

	rend = gtk_cell_renderer_text_new();
	col  = gtk_tree_view_column_new_with_attributes("Number", rend, "text",
							PATCH_COL_NUMBER_TEXT, NULL);
	gtk_tree_view_append_column(tv, col);

	col  = gtk_tree_view_column_new_with_attributes("Patch Name", rend, "text",
							PATCH_COL_NAME, NULL);
	gtk_tree_view_append_column(tv, col);
	sel = gtk_tree_view_get_selection(tv);
	g_signal_connect(G_OBJECT(sel), "changed",
			 G_CALLBACK(patch_selected), (gpointer)pch_file_indx);
    }
    gtk_list_store_clear(ls);
    if (!pch_file[pch_file_indx])
	return;

    for(i = 0; i < NUM_PATCHES; i++)
    {
	gtk_list_store_append(ls, &row);
	patch_update(pch_file_indx, i);
    }
}





/* Song handling */

void song_update(gint song)
{
    GtkListStore	*ls;
    GtkTreePath		*path;
    GtkTreeIter		 row;
    char		 buf[10];

    ls = GTK_LIST_STORE(gtk_tree_view_get_model(gui("SongTreeView")));
    path = gtk_tree_path_new_from_indices(song, -1);
    gtk_tree_model_get_iter(GTK_TREE_MODEL(ls), &row, path);

    gtk_list_store_set(ls, &row, SONG_COL_NUMBER, song, -1);
    sprintf(buf, "%2.2d", song+1);
    gtk_list_store_set(ls, &row, SONG_COL_NUMBER_TEXT, buf, -1);
    sprintf(buf, "%-6.6s", pts->song[song]->name);
    gtk_list_store_set(ls, &row, SONG_COL_NAME, buf, -1);
    gtk_tree_path_free(path);
}

gint song_selected_get(void)
{
    GtkTreeModel	*model;
    GtkTreeIter		 row1;
    gint		 song;

    /* Determine currently selected song number */
    if (!gtk_tree_selection_get_selected(
	   gtk_tree_view_get_selection(gui("SongTreeView")), &model, &row1))
	return -1;
    gtk_tree_model_get(model, &row1, SONG_COL_NUMBER, &song, -1);
    return song;
}

void song_select(gint song)
{
    GtkTreeView		*tv;
    GtkTreePath		*path;
    GtkTreeModel	*model;
    GtkTreeIter		 row;

    tv = gui("SongTreeView");
    model = gtk_tree_view_get_model(tv);
    path = gtk_tree_path_new_from_indices(song, -1);
    gtk_tree_model_get_iter(model, &row, path);
    gtk_tree_view_set_cursor(tv, path, NULL, FALSE);
    gtk_tree_path_free(path);
}

void song_move(gint direction)
{

    gint		 song1;
    gint		 song2;
    ps02_song_t		*temp;

    song1 = song_selected_get();
    if (song1 < 0)
	return;

    song2 = song1 + direction;
    if (song2 < 0 || song2 >= NUM_SONGS)
	return;

    /* Swap songs */
    temp = pts->song[song1];
    pts->song[song1] = pts->song[song2];
    pts->song[song2] = temp;
    pts->modified = TRUE;

    /* Update GUI */
    song_update(song1);
    song_update(song2);
    song_select(song2);
    return;
}

void song_selected(void)
{
    gint		song;
    gchar		name[7];

    measures_populate();

    song = song_selected_get();
    ignore_changed = TRUE;
    if (song < 0 || !pts)
    {
	gtk_entry_set_text(GTK_ENTRY(gui("SongName")), "");
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui("SongTempo")), 120);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui("SongRepeat")),
				     FALSE);
    } else {
	sprintf(name, "%-6.6s", pts->song[song]->name);
	gtk_entry_set_text(GTK_ENTRY(gui("SongName")), name);

	gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui("SongTempo")),
				  pts->song[song]->tempo);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui("SongRepeat")),
				     pts->song[song]->repeat);
    }
    ignore_changed = FALSE;
}


void song_list_populate(void)
{
    int			 i;
    GtkListStore	*ls;
    GtkTreeView		*tv;
    GtkTreeIter		 row;
    GtkTreeViewColumn	*col;
    GtkTreeSelection	*sel;
    GtkCellRenderer	*rend;

    tv = gui("SongTreeView");
    ls = GTK_LIST_STORE(gtk_tree_view_get_model(tv));
    if (ls == NULL)
    {
	ls = gtk_list_store_new(SONG_COLS, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING);
	gtk_tree_view_set_model(tv, GTK_TREE_MODEL(ls));

	rend = gtk_cell_renderer_text_new();
	col  = gtk_tree_view_column_new_with_attributes("Number", rend, "text",
							SONG_COL_NUMBER_TEXT, NULL);
	gtk_tree_view_append_column(tv, col);

	col  = gtk_tree_view_column_new_with_attributes("Song", rend, "text",
							SONG_COL_NAME, NULL);
	gtk_tree_view_append_column(tv, col);

	sel = gtk_tree_view_get_selection(tv);
	g_signal_connect(G_OBJECT(sel), "changed",
			 G_CALLBACK(song_selected), NULL);
    }
    gtk_list_store_clear(ls);
    if (!pts)
	return;

    for(i = 0; i < NUM_SONGS; i++)
    {
	gtk_list_store_append(ls, &row);
	song_update(i);
    }
}




/* Measure handling */

void measure_update(ps02_song_t *song, gint measure)
{
    GtkListStore	*ls;
    GtkTreePath		*path;
    GtkTreeIter		 row;
    char		 buf[80];
    int			 j;

    ls = GTK_LIST_STORE(gtk_tree_view_get_model(gui("RhythmTreeView")));
    path = gtk_tree_path_new_from_indices(measure, -1);
    gtk_tree_model_get_iter(GTK_TREE_MODEL(ls), &row, path);

    gtk_list_store_set(ls, &row, RHY_COL_MEASURE_NUMBER, measure, -1);
    sprintf(buf, "%2.2d", measure+1);
    gtk_list_store_set(ls, &row, RHY_COL_MEASURE, buf, -1);

    if (measure == song->measures)
    {
	gtk_list_store_set(ls, &row, RHY_COL_PATTERN, "END", -1);
    } else {
	sprintf(buf, "%-6.6s", pts->rhy_pattern[song->measure[measure].pattern]->name);
	gtk_list_store_set(ls, &row, RHY_COL_PATTERN, buf, -1);

	for (j = 0; j < song->measure[measure].beats; j++)
	{
	    unsigned char note = song->measure[measure].note[j];

	    if (note == 0xff)
		continue;

	    sprintf(buf, "%2s%3s ", note_root_name[note & 0x0f],
		    note_tone_name[(note & 0xf0) >> 4]);
	    gtk_list_store_set(ls, &row, RHY_COL_BEATS+j, buf, -1);
	}
    }

    gtk_tree_path_free(path);
}

void measures_selected(gint *beg, gint *end)
{
    GtkTreeSelection	*sel;
    GtkTreeModel	*model;
    GtkTreeIter		 row;
    GList		*list;

    sel = gtk_tree_view_get_selection(gui("RhythmTreeView"));
    list = gtk_tree_selection_get_selected_rows(sel, &model);

    if (g_list_first(list) != NULL)
    {
	if (beg)
	{
	    gtk_tree_model_get_iter(model, &row, g_list_first(list)->data);
	    gtk_tree_model_get(model, &row, RHY_COL_MEASURE_NUMBER, beg, -1);
	}

	if (end)
	{
	    gtk_tree_model_get_iter(model, &row, g_list_last(list)->data);
	    gtk_tree_model_get(model, &row, RHY_COL_MEASURE_NUMBER, end, -1);
	}

	gtk_tree_selection_select_range(sel, g_list_first(list)->data, g_list_last(list)->data);
    }

    g_list_foreach(list, (GFunc)gtk_tree_path_free, NULL);
    g_list_free(list);
}

void measures_selection_changed(void)
{
    measures_selected(NULL, NULL);
}


void measures_populate(void)
{
    int			 i;
    GtkListStore	*ls;
    GtkTreeView		*tv;
    GtkTreeIter		 row;
    GtkTreeViewColumn	*col;
    GtkCellRenderer	*rend;
    gint		 song;

    tv = gui("RhythmTreeView");
    ls = GTK_LIST_STORE(gtk_tree_view_get_model(tv));
    if (ls == NULL)
    {
	GtkTreeSelection	*sel;

	ls = gtk_list_store_new(MAX_BEATS_PER_MEASURE+3, G_TYPE_INT,
				G_TYPE_STRING, G_TYPE_STRING,
				G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
				G_TYPE_STRING, G_TYPE_STRING);
	gtk_tree_view_set_model(tv, GTK_TREE_MODEL(ls));

	rend = gtk_cell_renderer_text_new();
	col  = gtk_tree_view_column_new_with_attributes("Measure", rend, "text",
							RHY_COL_MEASURE, NULL);
	gtk_tree_view_append_column(tv, col);

	rend = gtk_cell_renderer_text_new();
	col  = gtk_tree_view_column_new_with_attributes("Pattern", rend, "text",
							RHY_COL_PATTERN, NULL);
	gtk_tree_view_append_column(tv, col);

	for (i = 0; i < MAX_BEATS_PER_MEASURE; i++)
	{
	    char buf[10];
	    sprintf(buf, "Beat %d", i+1);
	    rend = gtk_cell_renderer_text_new();
	    col  = gtk_tree_view_column_new_with_attributes(buf, rend, "text",
							    RHY_COL_BEATS+i, NULL);
	    gtk_tree_view_append_column(tv, col);
	}
	gtk_tree_view_set_rules_hint(tv, TRUE);
	sel = gtk_tree_view_get_selection(tv);
	gtk_tree_selection_set_mode(sel, GTK_SELECTION_MULTIPLE);
	g_signal_connect(G_OBJECT(sel), "changed",
			 G_CALLBACK(measures_selection_changed), NULL);
    }
    gtk_list_store_clear(ls);

    song = song_selected_get();

    if (!pts || song < 0)
	return;

    for(i = 0; i <= pts->song[song]->measures; i++)
    {
	gtk_list_store_append(ls, &row);
	measure_update(pts->song[song], i);
    }
}




/* Pattern handling */

void pattern_update(gint pattern)
{
    GtkListStore	*ls;
    GtkTreePath		*path;
    GtkTreeIter		 row;
    char		 buf[10];

    ls = GTK_LIST_STORE(gtk_tree_view_get_model(gui("PatternTreeView")));
    path = gtk_tree_path_new_from_indices(pattern, -1);
    gtk_tree_model_get_iter(GTK_TREE_MODEL(ls), &row, path);

    gtk_list_store_set(ls, &row, PATTERN_COL_NUMBER, pattern, -1);
    sprintf(buf, "%2.2d", pattern+1);
    gtk_list_store_set(ls, &row, PATTERN_COL_NUMBER_TEXT, buf, -1);
    sprintf(buf, "%-6.6s", pts->rhy_pattern[pattern]->name);
    gtk_list_store_set(ls, &row, PATTERN_COL_NAME, buf, -1);
    gtk_tree_path_free(path);
}

gint pattern_selected_get(void)
{
    GtkTreeModel	*model;
    GtkTreeIter		 row1;
    gint		 pattern;

    /* Determine currently selected pattern number */
    if (!gtk_tree_selection_get_selected(
	   gtk_tree_view_get_selection(gui("PatternTreeView")), &model, &row1))
	return -1;
    gtk_tree_model_get(model, &row1, PATTERN_COL_NUMBER, &pattern, -1);
    return pattern;
}

void pattern_selected(void)
{
    gint		pattern;
    gchar		name[7];

    //pattern_data_populate();

    pattern = pattern_selected_get();
    ignore_changed = TRUE;
    if (pattern < 0 || !pts)
    {
	gtk_entry_set_text(GTK_ENTRY(gui("PatternName")), "");
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui("PatternBeats")), 4);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui("PatternBars")), 1);
    } else {
	sprintf(name, "%-6.6s", pts->rhy_pattern[pattern]->name);
	gtk_entry_set_text(GTK_ENTRY(gui("PatternName")), name);

	gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui("PatternBeats")),
				  pts->rhy_pattern[pattern]->beats);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui("PatternBars")),
				  pts->rhy_pattern[pattern]->bars);
    }
    ignore_changed = FALSE;
}

void pattern_list_populate(void)
{
    int			 i;
    GtkListStore	*ls;
    GtkTreeView		*tv;
    GtkTreeIter		 row;
    GtkTreeViewColumn	*col;
    GtkTreeSelection	*sel;
    GtkCellRenderer	*rend;

    tv = gui("PatternTreeView");
    ls = GTK_LIST_STORE(gtk_tree_view_get_model(tv));
    if (ls == NULL)
    {
	ls = gtk_list_store_new(PATTERN_COLS, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING);
	gtk_tree_view_set_model(tv, GTK_TREE_MODEL(ls));

	rend = gtk_cell_renderer_text_new();
	col  = gtk_tree_view_column_new_with_attributes("Number", rend, "text",
							PATTERN_COL_NUMBER_TEXT, NULL);
	gtk_tree_view_append_column(tv, col);

	col  = gtk_tree_view_column_new_with_attributes("Pattern", rend, "text",
							PATTERN_COL_NAME, NULL);
	gtk_tree_view_append_column(tv, col);

	sel = gtk_tree_view_get_selection(tv);
	g_signal_connect(G_OBJECT(sel), "changed",
			 G_CALLBACK(pattern_selected), NULL);
    }
    gtk_list_store_clear(ls);
    if (!pts)
	return;

    for(i = 0; i < pts->rhy_pattern_cnt; i++)
    {
	gtk_list_store_append(ls, &row);
	pattern_update(i);
    }
}





gboolean lose_changes(void)
{
    if ((pts && pts->modified) ||
	(pch_file[0] && pch_file[0]->modified) ||
	(pch_file[1] && pch_file[1]->modified))
    {
	return modal_ask(TRUE, "Lose Changes", "Ok", "Cancel",
			 "You have made changes that are not yet saved.\n"
			 "Continue and lose these change?");
    }
    return TRUE;
}

gint on_quit(GtkMenuItem *mi, gpointer data)
{
    if (!lose_changes())
	return TRUE;

    gtk_main_quit();
    return FALSE;
}

void on_open(GtkMenuItem *mi, gpointer data)
{
    char file[PATH_MAX];

    on_close(NULL, NULL);
    modal_filesel("Open PTS file", file);
    open_files(file);
}

void on_save(GtkMenuItem *mi, gpointer data)
{
    if (!pts)
	return;

    PS02_pts_write(pts);
    pts->modified = FALSE;

    if (pch_file[0])
    {
	PS02_pch_write(pch_file[0]->F, pch_file[0]);
	pch_file[0]->modified = FALSE;
    }

    if (pch_file[1])
    {
	PS02_pch_write(pch_file[1]->F, pch_file[1]);
	pch_file[1]->modified = FALSE;
    }
}

void on_close(GtkMenuItem *mi, gpointer data)
{
    if (!pts)
	return;

    if (!lose_changes())
	return;

    PS02_pts_close(pts);
    if (pch_file[0])
	PS02_pch_close(pch_file[0]);
    if (pch_file[1])
	PS02_pch_close(pch_file[1]);
    pts = NULL;
    pch_file[0] = pch_file[1] = NULL;
    song_list_populate();
    song_selected();
    pattern_list_populate();
    patch_list_populate(0);
    patch_list_populate(1);
    gui_window_title(MainWindow, PROGRAM " "  VERSION);
}

void on_about(GtkMenuItem *mi, gpointer data)
{
    static const char gpl_terms[] = 
	"This program is free software; you can redistribute it and/or modify\n"
	"it under the terms of the GNU General Public License version 2 as \n"
	"published by the Free Software Foundation.\n \n"
	"This program is distributed in the hope that it will be useful, but\n"
	"WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
	"See the GNU General Public License for more details.";

    char buf[1024];
    sprintf(buf, 
	PROGRAM " " VERSION "\n"
	"Copyright (C) 2003-2004, Dwight Engen <dengen40@yahoo.com>\n\n"
	"%s", gpl_terms);

    modal_ask(FALSE, "About", "Ok", NULL, buf);
}

void on_repeat_changed(GtkWidget *w, gpointer data)
{
    gint songindx;

    if (!pts || ignore_changed)
	return;

    songindx = song_selected_get();
    if (songindx < 0)
	return;

    pts->song[songindx]->repeat =
	gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
    pts->modified = TRUE;
}

void on_songname_changed(GtkWidget *w, gpointer data)
{
    gint songindx;

    if (!pts || ignore_changed)
	return;

    songindx = song_selected_get();
    if (songindx < 0)
	return;

    strncpy(pts->song[songindx]->name, 
	   gtk_entry_get_text(GTK_ENTRY(gui("SongName"))), 6);
    song_update(songindx);
    pts->modified = TRUE;
}

void on_tempo_changed(GtkWidget *w, gpointer data)
{
    gint songindx;

    if (!pts || ignore_changed)
	return;

    songindx = song_selected_get();
    if (songindx < 0)
	return;

    pts->song[songindx]->tempo =
	gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(gui("SongTempo")));
    pts->modified = TRUE;
}



void on_patch_up(GtkWidget *w, gpointer data)
{
    if (!pts)
	return;

    patch_move(-1);
}

void on_patch_down(GtkWidget *w, gpointer data)
{
    if (!pts)
	return;

    patch_move(1);
}

void on_patch_copy(GtkWidget *w, gpointer data)
{
    gint pch_file_indx;
    gint pch;

    if (!pts)
	return;

    if (!patch_selected_get(&pch_file_indx, &pch))
	return;

    clippch = pch_file[pch_file_indx]->pch[pch];
}

void on_patch_paste(GtkWidget *w, gpointer data)
{
    gint pch_file_indx;
    gint pch;

    if (!pts)
	return;

    if (!patch_selected_get(&pch_file_indx, &pch))
	return;

    pch_file[pch_file_indx]->pch[pch] = clippch;
    patch_update(pch_file_indx, pch);
}

void on_patch_import(GtkWidget *w, gpointer data)
{
    char	 file[PATH_MAX];
    FILE	*IN;
    gint	 pch_file_indx;
    gint	 pch;

    if (!pts)
	return;

    if (!patch_selected_get(&pch_file_indx, &pch))
	return;

    modal_filesel("Import patch", file);
    if (file[0] == '\0')
	return;
    IN = fopen(file, "rb");
    if (!IN)
    {
	modal_ask(FALSE, "Import", "Ok", NULL, "Unable to open file %s", file);
	return;
    }

    PS02_patch_read(IN, &pch_file[pch_file_indx]->pch[pch]);
    pch_file[pch_file_indx]->modified = TRUE;
    fclose(IN);

    patch_update(pch_file_indx, pch);
    modal_ask(FALSE, "Import", "Ok", NULL, "Import complete");
}

void on_patch_export(GtkWidget *w, gpointer data)
{
    char	 file[PATH_MAX];
    FILE	*OUT;
    gint	 pch_file_indx;
    gint	 pch;

    if (!pts)
	return;

    if (!patch_selected_get(&pch_file_indx, &pch))
	return;

    modal_filesel("Export patch", file);
    if (file[0] == '\0')
	return;
    OUT = fopen(file, "w+b");
    if (!OUT)
    {
	modal_ask(FALSE, "Export", "Ok", NULL, "Unable to open file %s", file);
	return;
    }

    PS02_patch_write(OUT, &pch_file[pch_file_indx]->pch[pch]);
    fclose(OUT);
    modal_ask(FALSE, "Export", "Ok", NULL, "Export complete");
}



void on_song_up(GtkWidget *w, gpointer data)
{
    if (!pts)
	return;

    song_move(-1);
}

void on_song_down(GtkWidget *w, gpointer data)
{
    if (!pts)
	return;

    song_move(1);
}

void on_song_copy(GtkWidget *w, gpointer data)
{
    gint songindx;

    if (!pts)
	return;

    songindx = song_selected_get();
    if (songindx < 0)
	return;

    if (clipsong)
	PS02_song_free(clipsong);

    clipsong = PS02_song_copy(pts->song[songindx]);
}

void on_song_paste(GtkWidget *w, gpointer data)
{
    gint songindx;

    if (!pts)
	return;

    if (!clipsong)
	return;

    songindx = song_selected_get();
    if (songindx < 0)
	return;

    PS02_song_free(pts->song[songindx]);
    pts->song[songindx] = PS02_song_copy(clipsong);
    song_update(songindx);
    song_selected();
    pts->modified = TRUE;
}

void on_song_import(GtkWidget *w, gpointer data)
{
    char	 file[PATH_MAX];
    FILE	*IN;
    gint	 songindx;
    ps02_song_t	*song;

    if (!pts)
	return;

    songindx = song_selected_get();
    if (songindx < 0)
	return;

    modal_filesel("Import song", file);
    if (file[0] == '\0')
	return;
    IN = fopen(file, "rb");
    if (!IN)
    {
	modal_ask(FALSE, "Import", "Ok", NULL, "Unable to open file %s", file);
	return;
    }

    song = PS02_song_read(IN);
    fclose(IN);

    PS02_song_free(pts->song[songindx]);
    pts->song[songindx] = song;
    song_update(songindx);
    song_selected();
    pts->modified = TRUE;
    modal_ask(FALSE, "Import", "Ok", NULL, "Import complete");
}

void on_song_export(GtkWidget *w, gpointer data)
{
    char	 file[PATH_MAX];
    FILE	*OUT;
    gint	 songindx;

    if (!pts)
	return;

    songindx = song_selected_get();
    if (songindx < 0)
	return;

    modal_filesel("Export song", file);
    if (file[0] == '\0')
	return;
    OUT = fopen(file, "w+b");
    if (!OUT)
    {
	modal_ask(FALSE, "Export", "Ok", NULL, "Unable to open file %s", file);
	return;
    }

    PS02_song_write(OUT, pts->song[songindx]);
    fclose(OUT);
    modal_ask(FALSE, "Export", "Ok", NULL, "Export complete");
}

void on_song_export_ps04(GtkWidget *w, gpointer data)
{
    char	 file[PATH_MAX];
    gint	 songindx;
    ps04_seq_t	*seq;

    if (!pts)
	return;

    songindx = song_selected_get();
    if (songindx < 0)
	return;

    modal_filesel("Export song to .seq", file);
    if (file[0] == '\0')
	return;
    seq = PS04_seq_new();
    seq->F = fopen(file, "w+b");
    if (!seq->F)
    {
	modal_ask(FALSE, "Export", "Ok", NULL, "Unable to open file %s", file);
	return;
    }

    /* convert... */
    PS02_to_PS04(pts, seq, songindx);
    PS04_seq_write(seq);
    PS04_seq_close(seq);
    modal_ask(FALSE, "Export", "Ok", NULL, "Export complete");
}


void on_measure_copy(GtkWidget *w, gpointer data)
{
    if (!pts)
	return;

    measures_selected(&mcpy_beg, &mcpy_end);
}

void on_measure_paste(GtkWidget *w, gpointer data)
{
    gint song;

    if (!pts)
	return;

    measures_selected(&mcpy_dst, NULL);

    song = song_selected_get();
    PS02_song_copy_measures(pts->song[song], mcpy_beg, mcpy_end, mcpy_dst);
    pts->modified = TRUE;
    measures_populate();
}

void on_measure_delete(GtkWidget *w, gpointer user_data)
{
    gint song;

    if (!pts)
	return;

    measures_selected(&mcpy_beg, &mcpy_end);

    song = song_selected_get();
    PS02_song_delete_measures(pts->song[song], mcpy_beg, mcpy_end);
    pts->modified = TRUE;
    measures_populate();
}




/* General GUI handling routines */

void *gui(const gchar * widget_name)
{
    GtkWidget  *widget = MainWindow;
    GtkWidget  *parent, *found_widget;

    for (;;)
    {
	if (GTK_IS_MENU(widget))
	    parent = gtk_menu_get_attach_widget(GTK_MENU(widget));
	else
	    parent = widget->parent;
	if (!parent)
	    parent = g_object_get_data(G_OBJECT(widget), "GladeParentKey");
	if (parent == NULL)
	    break;
	widget = parent;
    }

    found_widget = (GtkWidget *) g_object_get_data(G_OBJECT(widget),
						   widget_name);
    if (!found_widget)
	g_warning("Widget not found: %s", widget_name);
    return found_widget;
}

typedef struct filesel_s
{
    GtkWidget	*win;
    char	*file;
} filesel_t;

static void modal_filesel_ok(GtkButton *button, filesel_t *fs)
{
    gchar *cd_dir;
    strcpy(fs->file, gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs->win)));

    /* chdir to the path given so the next selector starts there... */
    cd_dir = g_path_get_dirname(fs->file);
    chdir(cd_dir);
    g_free(cd_dir);

    gtk_main_quit();
}

static void modal_filesel_cancel(GtkButton *button, filesel_t *fs)
{
    fs->file[0] = '\0';
    gtk_main_quit();
}

static gboolean modal_filesel_destroy(GtkWidget *widget, GdkEvent *event,
				      filesel_t *fs)
{
    fs->file[0] = '\0';
    gtk_main_quit();
    return FALSE;
}

static void modal_filesel(const char *title, char *file)
{
    filesel_t fs;

    fs.file = file;
    fs.win = gtk_file_selection_new(title);
    gtk_window_set_modal(GTK_WINDOW(fs.win), TRUE);
    gtk_window_set_transient_for(GTK_WINDOW(fs.win), GTK_WINDOW(MainWindow));
    gtk_window_set_position(GTK_WINDOW(fs.win), GTK_WIN_POS_CENTER_ON_PARENT);

    gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(fs.win));
    gtk_signal_connect(GTK_OBJECT (fs.win),
		       "destroy_event", GTK_SIGNAL_FUNC(modal_filesel_destroy),
		       &fs);
    gtk_signal_connect(GTK_OBJECT (fs.win),
		       "delete_event", GTK_SIGNAL_FUNC(modal_filesel_destroy),
		       &fs);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fs.win)->ok_button),
		       "clicked", GTK_SIGNAL_FUNC(modal_filesel_ok),
		       &fs);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fs.win)->cancel_button),
		       "clicked", GTK_SIGNAL_FUNC(modal_filesel_cancel),
		       &fs);

    gtk_widget_show(fs.win);
    gtk_main();
    gtk_widget_destroy(fs.win);
    gui_sync();
}

static void on_modal_ask_ok(GtkButton *button, gpointer user_data)
{
    *(gboolean *)user_data = TRUE;
    gtk_main_quit();
}

static void on_modal_ask_cancel(GtkButton *button, gpointer user_data)
{
    *(gboolean *)user_data = FALSE;
    gtk_main_quit();
}

static gboolean on_modal_ask_destroy(GtkWidget *widget, GdkEvent *event,
			      gpointer user_data)
{
    *(gboolean *)user_data = FALSE;
    gtk_main_quit();
    return FALSE;
}

static gboolean modal_ask(gboolean can_cancel, const char *title,
			  const char *ok, const char *cancel,
			  const char *fmt,...)
{
    GtkWidget	*Win;
    GtkWidget	*MDLGVBox;
    GtkWidget	*MDLGLabel;
    GtkWidget	*separator;
    GtkWidget	*HButtonBox;
    GtkWidget	*Ok;
    GtkWidget	*Cancel;
    char	msg_buf[512];
    va_list	v_lst;
    gboolean	rc;

    Win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_transient_for(GTK_WINDOW(Win), GTK_WINDOW(MainWindow));
    gtk_window_set_position(GTK_WINDOW(Win), GTK_WIN_POS_CENTER_ON_PARENT);
    gtk_window_set_title(GTK_WINDOW (Win), title);
    gtk_window_set_modal(GTK_WINDOW (Win), TRUE);
    gtk_window_set_policy(GTK_WINDOW (Win), FALSE, TRUE, TRUE);

    MDLGVBox = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (MDLGVBox);
    gtk_container_add (GTK_CONTAINER (Win), MDLGVBox);
    gtk_container_set_border_width (GTK_CONTAINER (MDLGVBox), 2);

    va_start(v_lst, fmt);
    vsnprintf(msg_buf, sizeof(msg_buf)-1, fmt, v_lst);
    va_end(v_lst);

    MDLGLabel = gtk_label_new (msg_buf);
    gtk_widget_show (MDLGLabel);
    gtk_box_pack_start (GTK_BOX (MDLGVBox), MDLGLabel, TRUE, TRUE, 0);
    gtk_label_set_justify (GTK_LABEL (MDLGLabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment (GTK_MISC (MDLGLabel), 0.5, 0.10);

    separator = gtk_hseparator_new ();
    gtk_widget_show (separator);
    gtk_box_pack_start (GTK_BOX (MDLGVBox), separator, FALSE, FALSE, 4);

    HButtonBox = gtk_hbutton_box_new ();
    gtk_widget_show (HButtonBox);
    gtk_box_pack_end (GTK_BOX (MDLGVBox), HButtonBox, FALSE, FALSE, 0);
    gtk_button_box_set_layout (GTK_BUTTON_BOX (HButtonBox), GTK_BUTTONBOX_END);
    gtk_button_box_set_spacing (GTK_BUTTON_BOX (HButtonBox), 5);

    Ok = gtk_button_new_with_label (ok);
    gtk_widget_show (Ok);
    gtk_container_add (GTK_CONTAINER (HButtonBox), Ok);
    GTK_WIDGET_SET_FLAGS (Ok, GTK_CAN_DEFAULT);

    if (can_cancel)
    {
	Cancel = gtk_button_new_with_label (cancel);
	gtk_widget_show (Cancel);
	gtk_container_add (GTK_CONTAINER (HButtonBox), Cancel);
	GTK_WIDGET_SET_FLAGS (Cancel, GTK_CAN_DEFAULT);

	gtk_signal_connect (GTK_OBJECT (Cancel), "clicked",
			    GTK_SIGNAL_FUNC (on_modal_ask_cancel),
			    &rc);
    }

    gtk_signal_connect (GTK_OBJECT (Win), "destroy_event",
                        GTK_SIGNAL_FUNC (on_modal_ask_destroy),
                        &rc);
    gtk_signal_connect (GTK_OBJECT (Win), "delete_event",
                        GTK_SIGNAL_FUNC (on_modal_ask_destroy),
                        &rc);
    gtk_signal_connect (GTK_OBJECT (Ok), "clicked",
                        GTK_SIGNAL_FUNC (on_modal_ask_ok),
                        &rc);

    gtk_widget_grab_default (Ok);
    gtk_widget_show(Win);
    gtk_main();
    gtk_widget_destroy(Win);

    return rc;
}



static void open_files(const char *file)
{
    gchar *dir;
    gchar *pch_file_name;

    dir = g_path_get_dirname(file);

    pts = PS02_pts_open(file);

    /* Preset patches */
    pch_file_name = g_build_filename(dir, "Preset.pch", NULL);
    pch_file[0] = PS02_pch_open(pch_file_name);
    g_free((gpointer)pch_file_name);

    /* User patches */
    pch_file_name = g_build_filename(dir, "User.pch", NULL);
    pch_file[1] = PS02_pch_open(pch_file_name);
    g_free((gpointer)pch_file_name);

    song_list_populate();
    pattern_list_populate();
    patch_list_populate(0);
    patch_list_populate(1);

    gui_window_title(MainWindow, PROGRAM " " VERSION " - %s", file);
    g_free((gpointer)dir);
}


int main(int argc, char *argv[])
{
    gtk_set_locale();
    gtk_init(&argc, &argv);

    MainWindow = create_MainWindow();
    gtk_widget_show(MainWindow);

    /* open file specified on cmd line */
    if (argc > 1)
	open_files(argv[1]);
    else
	gui_window_title(MainWindow, PROGRAM " "  VERSION);

    gtk_main();
    return 0;
}

void
on_patternname_changed                 (GtkEditable     *editable,
                                        gpointer         user_data)
{

}


void
on_beats_changed                       (GtkEditable     *editable,
                                        gpointer         user_data)
{

}


void
on_bars_changed                        (GtkEditable     *editable,
                                        gpointer         user_data)
{

}


void
on_pattern_copy                        (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_pattern_paste                       (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_pattern_import                      (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_pattern_export                      (GtkButton       *button,
                                        gpointer         user_data)
{

}
